/***********************************************************************/
/*                                                                     */
/*                        OCamlP3l -- Example                          */
/*                                                                     */
/*             projets Estime & Cristal, INRIA-Rocquencourt            */
/*                                                                     */
/*  Copyright 2003-2007 INRIA.                                         */
/*  Institut National de Recherche en Informatique et en Automatique.  */
/*  All rights reserved. This file is distributed only by permission.  */
/*                                                                     */
/*  Francois Clement <Francois.Clement@inria.fr>                       */
/*  Pierre Weis <Pierre.Weis@inria.fr>                                 */
/*                                                                     */
/***********************************************************************/

/* $Id: poisson.c,v 1.1 2007/01/17 23:36:00 fclement Exp $ */

/*
 * This program implements the Dirichlet-to-Dirichlet operator S_f for the
 * solution of the 2D Poisson equation on a rectangular box domain using an
 * additive Schwarz algorithm, i.e. domain decomposition with overlapping and
 * Dirichlet transmission conditions.
 * 
 *              +--------------------------+
 *              |  o                    o  |
 *              |oooooooooooooooooooooooooo|
 *              |  o                    o  |
 *              |  o                    o  |
 *              |  o                    o  |       - + |  Gamma
 *              |  o                    o  |       o      ~Gamma
 *              |  o       Omega        o  |
 *              |  o                    o  |
 *              |  o                    o  |
 *              |  o                    o  |
 *              |  o                    o  |
 *              |  o                    o  |
 *              |oooooooooooooooooooooooooo|
 *              |  o                    o  |
 *              +--------------------------+
 *              
 * Let Omega be a subdomain, Gamma its boundary, and ~Gamma the trace of its
 * neighbors. Given right-hand sides f on Omega and g on Gamma, let u be the
 * solution of the Poisson equation:
 *                   - Laplacian(u) = f on Omega
 *                               u  = g on Gamma.
 * For a given f, S_f is the operator:
 *                       g on Gamma -> ~g on ~Gamma
 * where ~g is the trace of u on ~Gamma.
 * 
 * The program first reads on its standard input a 2D regular computational
 * grid, then it enters an infinite loop. For each Dirichlet condition g read
 * on its standard input:
 * - it computes using a simple SOR solver the solution u of the Poisson
 *   equation for a fixed right-hand side f,
 * - it optionnally prints the solution u into a file,
 * - it writes ~g on its standard output.
 * 
 * Written from a C program by Arnaud Vodicka.
 * 
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#define MAX(a,b) (a>=b ? a : b)
#define PI 3.1415926535589793
#define EPS 0.5

/* Choice of the right-hand side f */
float f(float x,float y) {
  return 2*(x*(1. - x)
            + y*(1. - y));
}

/* The corresponding solution u */
float u(float x,float y) {
  return x*(1. - x)*y*(1. - y);
}

/* Allocation of arrays */

float *array1d_alloc(int len) {
  float *x;
  
  x = (float *)calloc((unsigned)len,sizeof(float));
  if (x==NULL) {fprintf(stderr,"unable to allocate memory"); exit(1); }
  return x;
}

float **array2d_alloc(int leni,int lenj) {
  int i;
  float **x;
  
  x = (float **)calloc((unsigned)leni,sizeof(float *));
  if (x==NULL) {fprintf(stderr,"unable to allocate memory"); exit(1); }
  for (i=0; i<leni; i++) {x[i] = array1d_alloc(lenj); }
  return x;
}

/* Computing coordinates, source and solution */

float *coordinates(int nx,float x0,float dx) {
  float *x;
  int i;
  
  x = array1d_alloc(nx+1);
  for (i=0; i<=nx; i++) {x[i] = x0 + i*dx; }
  return x;
}

float *display_coordinates(int nx,int px,float dx,float *x) {
  float *xp;
  int i;
  
  xp = array1d_alloc(nx+1);
  for (i=0; i<=nx; i++) {
    xp[i] = x[i];
    if (i<=2*px) {xp[i] += (2.*px-i)/(2.*px)*EPS*dx; }
    if (i>=nx-2*px) {xp[i] -= (i-nx+2.*px)/(2.*px)*EPS*dx; }
  }
  return xp;
}

float **source(int nx,int ny,float *x,float *y) {
  float **ff;
  int i,j;
  
  ff = array2d_alloc(nx+1,ny+1);
  for (i=0; i<=nx; i++) {
    for (j=0; j<=ny; j++) {
      ff[i][j] = f(x[i],y[j]);
    }
  }
  return ff;
}

float **solution(int nx,int ny,float *x,float *y) {
  float **uu;
  int i,j;
  
  uu = array2d_alloc(nx+1,ny+1);
  for (i=0; i<=nx; i++) {
    for (j=0; j<=ny; j++) {
      uu[i][j] = u(x[i],y[j]);
    }
  }
  return uu;
}

int secure_scanff(float *f) {
  int ret_code;
  ret_code = scanf("%f",f);
  if (ret_code == EOF || ret_code < 1) { exit (2); }
  return 0;
}

int secure_scanfd(int *d) {
  int ret_code;
  ret_code = scanf("%d",d);
  if (ret_code == EOF || ret_code < 1) { exit (2); }
  return 0;
}

int secure_scanfs(char s[255]) {
  int ret_code;
  ret_code = scanf("%256s",s);
  if (ret_code == EOF) { exit (0); }
  if (ret_code < 1) { exit (2); }
  return 0;
}

int main() {
  
  int i,j,nx,ny,px,py,iter,itermax;
  float x0,y0,lx,ly;
  float dx,dy,dx2,dy2,mdx2,mdy2;
  float eta,w,alpha,beta,**uold,**unew;
  float *x,*y,*xp,*yp,**ff,**uu;
  float rel_stag,rel_err,max_uu;
  char result[255];
  FILE *fp;
  
  /****************************************/
  /* READ THE COMPUTATIONAL GRID ON STDIN */
  /****************************************/
  
  secure_scanff(&x0); /* x of the reference corner */
  secure_scanff(&y0); /* y of the reference corner */
  secure_scanff(&lx); /* size of the domain in the x-direction */
  secure_scanff(&ly); /* size of the domain in the y-direction */
  secure_scanfd(&nx); /* number of cells in the x-direction */
  secure_scanfd(&ny); /* number of cells in the y-direction */
  secure_scanfd(&px); /* half-overlapping in the x-direction */
  secure_scanfd(&py); /* half-overlapping in the y-direction */
  secure_scanfd(&itermax); /* max number of iterations */
  
  /******************/
  /* INITIALIZATION */
  /******************/
  
  dx = lx/nx; dy = ly/ny;
  dx2 = dx*dx; dy2 = dy*dy;
  mdx2 = 1./dx2; mdy2 = 1./dy2;
  
  /* Tuning the numerical scheme */
  eta = 1. - 2.*sin(PI/(2.*nx))*sin(PI/(2.*ny));
  w = 2./(1. + sqrt(1. - eta*eta));
  alpha = 0.5/(mdx2 + mdy2)*w;
  beta = 1. - w;
  /* Allocating arrays for the two-step numerical scheme */
  uold = array2d_alloc(nx+1,ny+1);
  unew = array2d_alloc(nx+1,ny+1);
  
  /* Computing coordinates, source and solution */
  x = coordinates(nx,x0,dx);
  y = coordinates(ny,y0,dy);
  xp = display_coordinates(nx,px,dx,x);
  yp = display_coordinates(ny,py,dy,y);
  ff = source(nx,ny,x,y);
  uu = solution(nx,ny,x,y);
  
  /*****************/
  /* INFINITE LOOP */
  /*****************/
  
  while (1) {
    
    /***********************************************/
    /* READ THE DIRICHLET CONDITION g ON STDIN     */
    /* Be careful, each corner is given two times! */
    /***********************************************/
    
    /* Name of the file to save this iteration */
    /* "none" means no saving, EOF means end of computation */
    secure_scanfs(result);
    
    /* West side */
    for (j=0; j<=ny; j++) {
      secure_scanff(&uold[0][j]);
      unew[0][j] = uold[0][j];
    }
    /* East side */
    for (j=0; j<=ny; j++) {
      secure_scanff(&uold[nx][j]);
      unew[nx][j] = uold[nx][j];
    }
    /* South side */
    for (i=0; i<=nx; i++) {
      secure_scanff(&uold[i][0]);
      unew[i][0] = uold[i][0];
    }
    /* North side */
    for (i=0; i<=nx; i++) {
      secure_scanff(&uold[i][ny]);
      unew[i][ny] = uold[i][ny];
    }
    
    /************************************/
    /* SOLVE {Laplacian(u) = f on Omega */
    /*       {          u  = g on Gamma */
    /************************************/
    
    iter = 0;
    
    /* Complexity = 9*(nx-1)*(ny-1)*itermax flop */
    do { 
      for (i=1; i<=nx-1; i++) {
        for(j=1; j<=ny-1; j++) {
          uold[i][j] = unew[i][j];
        }
      }
      
      for (i=1; i<=nx-1; i++) {
        for (j=1; j<=ny-1; j++) {
          unew[i][j] = alpha*(mdx2*(uold[i+1][j] + unew[i-1][j])
                              + mdy2*(uold[i][j+1] + unew[i][j-1])
                              + ff[i][j])
                     + beta*uold[i][j];
        }
      }
      
      iter++;
      
    } while (iter<itermax);
    
    /* Compute the relative stagnation and the relative error */
    rel_stag = 0.; rel_err = 0.; max_uu = 0.;
    
    for (i=0; i<=nx; i++) {
      for (j=0; j<=ny; j++) {
        if (fabs(uold[i][j])>0) {
          rel_stag = MAX(rel_stag,
                         fabs(unew[i][j]
                              - uold[i][j])/fabs(uold[i][j]));
        }
        max_uu = MAX(max_uu,fabs(uu[i][j]));
        rel_err = MAX(rel_err,fabs(uu[i][j] - uold[i][j]));
      }
    }
    rel_err /= max_uu;
    
    /************************************/
    /* PRINT THE SOLUTION u INTO A FILE */
    /************************************/
    
    if (strcmp(result,".")) {
      fp = fopen(result,"w");
      if (fp==NULL) {
        fprintf(stderr,"unable to open file %s\n",result); exit(3);
      }
      for (i=0; i<=nx; i++) {
        for (j=0; j<=ny; j++) {
          fprintf(fp,"%f %f %f\n",xp[i],yp[j],unew[i][j]);
        }
      }
      fclose(fp);
    }
    
    /**********************************/
    /* PRINT THE TRACE OF u ON STDOUT */
    /**********************************/
    
    printf("%f %f\n",rel_stag,rel_err);
    
    /* West side */
    for (j=0; j<=ny; j++) {
      printf("%f\n",unew[2*px][j]);
    }
    /* East side */
    for (j=0; j<=ny; j++) {
      printf("%f\n",unew[nx-2*px][j]);
    }
    /* South side */
    for (i=0; i<=nx; i++) {
      printf("%f\n",unew[i][2*py]);
    }
    /* North side */
    for (i=0; i<=nx; i++) {
      printf("%f\n",unew[i][ny-2*py]);
    }
    fflush(stdout);
    
  } /* end while */
  
  return 0;
  
} /* end main */
